import enum
import pytest
import socket

from si_tests import settings, logger
from si_tests.managers import bootstrap_manager
from si_tests.managers import pdns_manager
from si_tests.managers.kaas_manager import Manager
from si_tests.utils import certs

LOG = logger.logger
bootstrap = bootstrap_manager.BootstrapManager.get_si_config_bootstrap_manager()


class TLSServices(enum.Enum):
    """Prefixes for interfaces that do not require some types of checks
    Items:
        service_namespace: Namespace of the service
        service_name:      Name of the service
        dns_records:       List of DNS records to add into the Cluster DNS zone for this service
        cluster_tls_name:  Name of the TLS key for the Cluster.spec.providerSpec.value.tls dictionary
        description:       Item description for logging
        for_management:    Boolean, 'True' for testing the service on Management cluster, 'False' is to skip
        for_regional:      Boolean, 'True' for testing the service on Region cluster, 'False' is to skip
        for_child:         Boolean, 'True' for testing the service on Child cluster, 'False' is to skip
        for_mosk:          Boolean, 'True' for testing the service on MOSK cluster, 'False' is to skip
    """
    alerta = ("stacklight", "iam-proxy-alerta", "alerta.stacklight", "iamProxyAlerta",
              "Alerta (StackLight)", True, True, True, True)
    alertmanager = ("stacklight", "iam-proxy-alertmanager", "alertmanager.stacklight", "iamProxyAlertManager",
                    "Alert Manager (StackLight)", True, True, True, True)
    grafana = ("stacklight", "iam-proxy-grafana", "grafana.stacklight", "iamProxyGrafana",
               "Grafana (StackLight)", True, True, True, True)
    kibana = ("stacklight", "iam-proxy-kibana", "kibana.stacklight", "iamProxyKibana",
              "Kibana (StackLight)", True, True, True, True)
    prometheus = ("stacklight", "iam-proxy-prometheus", "prometheus.stacklight", "iamProxyPrometheus",
                  "Prometheus (StackLight)", True, True, True, True)
    cache = ("kaas", "mcc-cache", "cache.kaas", "cache",
             "MCC Cache", True, True, False, False)
    ui = ("kaas", "kaas-kaas-ui", "ui.kaas", "ui",
          "MCC UI", True, False, False, False)
    keycloak = ("kaas", "iam-keycloak-http", "keycloak.kaas", "keycloak",
                "MCC Keycloak", True, False, False, False)
    mke = (None, None, "mke", "mke",
           "MKE API", True, True, True, False)

    def __init__(self, service_namespace, service_name, dns_record, cluster_tls_name, description,
                 for_management, for_regional, for_child, for_mosk):
        self.service_namespace = service_namespace
        self.service_name = service_name
        self.dns_record = dns_record
        self.cluster_tls_name = cluster_tls_name
        self.description = description
        self.for_management = for_management
        self.for_regional = for_regional
        self.for_child = for_child
        self.for_mosk = for_mosk


def save_pem_artifact(filename, pem_data):
    with open(f'{settings.ARTIFACTS_DIR}/{filename}', mode='w') as f:
        f.write(pem_data)


def tls_services():
    """Prepare set of parameters to test each service with updated TLSConfig

    1. Generate CA certificate that will be used for all the services in this cluster
    2. Fill 'yield_data' with the services related only to the current cluster type
       (management, region, child or mosk)
    3. For each related service, create DNS record in the PDNS on seed node, using
       the IP address from the kubernetes 'services' object.
       Exception: MKE IP address, which is taken from 'loadBalancerHost' after resolving
       it from a hostname (if there is a hostname instead of IP)
    """

    cluster_name = settings.TARGET_CLUSTER
    cluster_namespace = settings.TARGET_NAMESPACE
    build = f"build{settings.JENKINS_BUILD_NUMBER}"
    zone_name = f"{build}.{cluster_name}.{cluster_namespace}.{settings.SEED_PDNS_LOCAL_ZONE}"

    LOG.info("Generate CA certificate for the specified TARGET_CLUSTER public services")
    ca_pem, ca_key_pem, ca_path = bootstrap._generate_ca_cert(ca_path=zone_name)
    save_pem_artifact('ca.pem', ca_pem)
    save_pem_artifact('ca_key.pem', ca_key_pem)

    LOG.info("Create DNS records in PDNS Authoritative server for the specified TARGET_CLUSTER public services")
    pdns = pdns_manager.PDNSManager(bootstrap_manager=bootstrap)
    pdns.create_zone(zone_name)

    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    ns = kaas_manager.get_namespace(cluster_namespace)
    cluster = ns.get_cluster(cluster_name)
    # Save KUBECONFIG to artifacts to work with job artifacts from the pipeline scenario
    certs.update_seed_kubeconfig(bootstrap=bootstrap, cluster=cluster)

    is_management = cluster.is_management
    is_regional = cluster.is_regional
    is_child = cluster.is_child
    is_mosk = cluster.is_mosk
    k8sclient = cluster.k8sclient
    yield_data = []
    for s in TLSServices:
        if not any([is_management and s.for_management,
                    is_regional and s.for_regional,
                    # if MOSK, then skip some services
                    is_mosk and s.for_mosk,
                    (not is_mosk) and is_child and s.for_child]):
            LOG.warning(f"Skip testing service '{s.description}' for the cluster '{cluster_namespace}/{cluster_name}'")
            continue

        if s.name == 'mke':
            # get IP for MKE from Cluster loadBalancerHost
            try:
                ip = socket.gethostbyname(cluster.load_balancer_ip)
            except socket.gaierror as e:
                raise Exception(f"Invalid Cluster loadbalancer host '{cluster.load_balancer_ip}': {e}")
        else:
            # get IP from service
            if not k8sclient.services.present(name=s.service_name, namespace=s.service_namespace):
                raise Exception(f"Service {s.service_namespace}/{s.service_name} not found in the Cluster "
                                f"{cluster_namespace}/{cluster_name}")
            service = k8sclient.services.get(name=s.service_name, namespace=s.service_namespace)
            ip = service.get_external_ip()
            if not ip:
                raise Exception(f"Service {s.service_namespace}/{s.service_name} don't contain external IP address "
                                f"in the Cluster {cluster_namespace}/{cluster_name}")

        LOG.info(f"Creating PDNS records for '{s.name}' with ip '{ip}'")
        pdns.add_record(zone_name, s.dns_record, 'A', ip)

        yield_data.append({
            'name': s.name,
            'dns_name': f"{s.dns_record}.{zone_name}",
            'ip': ip,
            'cluster_tls_name': s.cluster_tls_name,
            'description': s.description,
            'ca': (ca_pem, ca_key_pem, ca_path),
            'is_management': is_management,
            'is_regional': is_regional,
            'is_child': is_child,
            'is_mosk': is_mosk,
        })

    pdns.show_zone(zone_name)
    return yield_data


tls_services_params = list(tls_services())


@pytest.mark.usefixtures("introspect_child_target_objects")
@pytest.mark.parametrize("tls_service", tls_services_params,
                         ids=[f"TLS_SERVICE={x['name']}" for x in tls_services_params])
def test_tlsconfig_for_public_services(kaas_manager, tls_service, show_step):
    """Test TLSConfig with custom DNS names and certificates for cluster public services
    Use only with RUN_ON_REMOTE=True, to access the DNS resolver on the seed node

    Scenario:
        1. Generate certificate for public service using the created DNS records
        2. Apply the certificate using TLSConfig object
        3. Check related Clusters readiness
        4. Check that certificate on the Machines in the current region actually contains the expected DNS name
        5. Check that the new CA is applied for the kubeconfig and helm-controller secrets
        6. Enable RUN_ON_REMOTE=True in SI_CONFIG, so all further tests could resolve the DNS names on seed node PDNS
        # 4. Remove custom TLSConfig objects
    """
    cluster_name = settings.TARGET_CLUSTER
    cluster_namespace = settings.TARGET_NAMESPACE
    ns = kaas_manager.get_namespace(cluster_namespace)
    cluster = ns.get_cluster(cluster_name)

    show_step(1)
    LOG.info(f"Generate certificate for the service '{tls_service['description']}'")
    ca_pem, ca_key_pem, ca_path = tls_service['ca']

    if (tls_service['name'] == 'mke' and cluster.workaround.prodx_34281() is True):
        # WA for PRODX-34281: generate MKE certificate with IP instead of DNS in case of BM/EM2/vsphere providers
        # Remove this WA after PRODX-34281 is completed
        cert_pem, key_pem = bootstrap._generate_app_cert(
            ips=[tls_service['ip']],
            ca_path=ca_path,
            cert_dir=tls_service['name'])
        # Apply certificate for IP instead of DNS name
        tls_service['dns_name'] = tls_service['ip']
    else:
        # Use only DNS name for the certificate, to test the following case:
        # - if there is any client that is still using the service IP address to access the service,
        #   then it must fail because certificate requires DNS name, not IP.
        cert_pem, key_pem = bootstrap._generate_app_cert(
            dns_names=[tls_service['dns_name']],
            ca_path=ca_path,
            cert_dir=tls_service['name'])
    save_pem_artifact(f"cert_{tls_service['name']}.pem", cert_pem)
    save_pem_artifact(f"cert_{tls_service['name']}_key.pem", key_pem)

    # Update KUBECONFIG_ARTIFACT if MKE certificate will be updated for Management or Regional cluster
    if tls_service['name'] == 'mke':
        # if cluster.is_management or cluster.is_regional:
        # Use KUBECONFIG with disabled TLS verify, to allow working with certificate hostnames which
        # are not resolving from the Jenkins agents
        kubeconfig_path = certs.update_seed_kubeconfig(bootstrap=bootstrap, cluster=cluster, skip_tls_verify=True)
        if cluster.is_management:
            # If MKE certificate will be updated for Management, then re-initialize
            # 'kaas_manager' object using the KUBECONFIG with disabled TLS verification
            # IMPORTANT: use this KUBECONFIG for all next tests in the pipeline
            #            in case if these tests are running without RUN_ON_REMOTE flag
            #            and MKE certificate uses the DNS name, which is created on seed node PDNS
            #            and is not resolved from other nameservers in the network.
            kaas_manager.api.update_kubeconfig(kubeconfig_path)

    # Update the certificate
    show_step(2)
    cert_manager = certs.CertManager(kaas_manager)
    cert_manager.apply_cert_for_app(
        app=tls_service['name'],
        cluster=cluster,
        hostname=tls_service['dns_name'],
        cert_pem=cert_pem,
        key_pem=key_pem,
        ca_pem=ca_pem,
        app_tlsspec=tls_service['cluster_tls_name'])

    # 1. If MKE is updated on Management - check all clusters readiness, and agent/helmcontroller
    #      in clusters of the current region, update kubeconfig
    #    If MKE is updated on Region - check region and it's clusters readiness, and agent/helmcontroller
    #      in clusters of the current region, update kubeconfig
    #    If MKE is updated on Child - check child readiness.
    # 2. If Cache - check readiness as in #1 (except childs)
    # 3. If Keycloak - check readiness as in #1 for Management
    show_step(3)

    # Management cluster to check readiness
    check_clusters = []
    if cluster.is_management:
        if tls_service['name'] in ['mke', 'cache', 'keycloak']:
            # Check cluster readiness after MKE, Cache or Keycloak certificate changed
            check_clusters.append(cluster)  # Management cluster
            check_clusters.extend(kaas_manager.get_regional_clusters())  # All Region clusters
            check_clusters.extend(kaas_manager.get_child_clusters())     # All Child clusters
        else:
            # Check only Management cluster for all other services
            check_clusters.append(cluster)  # Management cluster

    # Region cluster to check readiness
    elif cluster.is_regional:
        if tls_service['name'] in ['mke', 'cache']:
            # Check cluster readiness after MKE, Cache or Keycloak certificate changed
            check_clusters.append(cluster)  # Region cluster
            check_clusters.append(kaas_manager.get_mgmt_cluster())  # Management cluster
            # Child clusers from current MCC region
            check_clusters.extend(kaas_manager.get_child_clusters_in_region(cluster.region_name))
        else:
            # Check only Region cluster for all other services
            check_clusters.append(cluster)  # Region cluster

    # Child cluster to check readiness
    else:
        check_clusters.append(cluster)  # Child cluster

    # Check clusters readiness
    for check_cluster in check_clusters:
        check_cluster.check.check_cluster_readiness()

    show_step(4)
    # Check that certificate on the host actually contains the expected DNS name
    cluster.check.check_cert_conversation(tls_service['dns_name'])

    show_step(5)
    if cluster.is_management or cluster.is_regional:
        # If Region or Management TLS certificates were changed, then it may affect all the clusters in the region
        if tls_service['name'] == 'mke':
            target_clusters = [cluster]
            target_clusters.extend(kaas_manager.get_child_clusters_in_region(cluster.region_name))
            for target_cluster in target_clusters:
                LOG.info(f"Checking cluster {target_cluster.name} certificate")
                target_cluster.check.validate_agent_kubeconfig(ca=ca_pem)
                target_cluster.check.validate_helm_controller_secret(ca=ca_pem)
        else:
            LOG.info("MKE certificates was not updated, skipping check for updated CA in the secrets")
    else:
        LOG.info("Skip Management/Region CA checks for a Child cluster")

    show_step(6)
    kaas_manager.si_config.update_run_on_remote(run_on_remote=True)
